
local object = require("object")
local canBuildFunctions = require("canBuild")
-- Settings for Imperialism Version III

-- canBuildParameters
--      Three tables, one for unitTypes, one for Improvements, one for Wonders
--      absent entry means use the defaultCanBuild function
-- canBuildObjectType[item.id]= {
--      .forbiddenTribes = {[tribeID]=bool}
--          if canBuildObjectType[item.id].forbiddenTribes[tribeID] is true, then the tribe with
--          tribeID can't build item, false or nil/absent means it can
--
--      .forbiddenMaps = {[0] = bool,[1]=bool,[2]=bool,[3]=bool}
--          if canBuildObjectType[item.id].forbiddenMaps[mapCityIsOn] = true, then city can't build the item
--              false or nil means it can
--          absent means all maps are allowed
--      .location = {xCoord,yCoord} or {xCoord,yCoord,zCoord} or tileObject or cityObject or integer or function(tileObject)-->boolean or table of these kinds of objects
--          {xCoord,yCoord} if the city is located at (xCoord,yCoord) on any map, it can build the object
--          {xCoord,yCoord,zCoord} means the city must be located at those coordinates to build the object
--          tileObject means the city must be located at that tile
--          cityObject means the city must be that city
--          function means object can be built if function(city.location) returns true 
--          (and all other conditions are met), and can't be built otherwise
--          table of these things means that each entry in the table is checked, and if any one of them means the object can be built, then it can be built
--          absent means the object can be built at any location
--          (Note: Can't use integers to match city id, since code can't distinguish between several cities and a coordinate triple)
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .forbiddenLocation= {xCoord,yCoord} or {xCoord,yCoord,zCoord} or tileObject or cityObject or function(tileObject)-->boolean or table of these kinds of objects
--              see location details, except that a match in forbidden location prevents the item from being buitl
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .allImprovements = improvementObject or wonderObject or table of improvementObjects and wonderObjects
--          The city must have all improvements/wonders in the table to build the object
--          absent means no improvements needed (in this section)
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .someImprovements = improvementObject or wonderObject or table of improvementObjects and wonderObjects
--          The city must have some number of objects in the table to build the item in question.  The exact
--          number is governed by the next key
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .numberOfImprovements = integer
--          tells how many of the 'someImprovements' are needed to build the item
--          absent means ignore .someImprovements
--      .forbiddenImprovements = improvementObject or wonderObject or table of improvementObjects and wonderObjects
--          The city must have none of the improvements/wonders in the table to build the object
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .allTechs = technologyObject or table of technologyObjects
--          the civ must have all the technologies in the table to build the object
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .someTechs = technologies or table of technologyObjects
--          the civ must have some of the technologies in the table to build the object
--          the number is governed by the next key
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .numberOfTechs = integer
--          tells how many of the 'someTechs' are needed for the object to be built
--      .forbiddenTechs = technologyObject or table of technologyObjects
--          the civ must not have any of the technologies in the table to build the object
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .allFlagsMatch = {[flagKey]=boolean}
--          the city can only build the item if all the flags for the flagKeys in the table have the  
--          corresponding value
--          absent flag key (including [flagKey]=nil) doesn't affect production
--          absent means no restriction
--      .someFlagsMatch = {[flagKey]=boolean}
--          the city can only build the item if all the flags for the flagKeys in the table have the  
--          corresponding value
--          absent flag key (including [flagKey]=nil) doesn't affect production
--          absent means no restriction
--      .numberOfFlags = integer
--          tells how many of the 'someFlags' are needed for the object to be built
--      .minimumPopulation = integer
--          the city must have at least this many citizens to build the item
--          absent means 0
--      .maximumPopulation = integer 
--          the city can have at most this many citizens to build the item
--          absent means no maximum population
--      .earliestTurn = integer
--          item can't be built before this turn
--          absent means no restriction
--      .latestTurn = integer
--          item can't be built after this turn
--          absent means no restriction
--      .allWonders = wonderObject or table of wonderObject
--          the city's tribe must have all these wonders in order to build the item
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .someWonders = wonderObject or table of wonderObject
--          the city's tribe must have some of the wonders in order to build the item
--          the number is governed by the next key
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .numberOfWonders = integer
--          tells how many of 'someWonders' the tribe must have to build the item
--          absent means no restriction
--
--
--      .overrideDefaultBuildFunction = boolean or nil
--          if true, the in game function for determining if the item can be built is ignored
--          for improvements and wonders, a check will be made if the item has already been built
--          false, nil, absent means the game's regular conditions must also be met
--      .forbiddenAlternateProduction = unitTypeObject or imprvementObject or wonderObject or table of these objects
--          the city can only build the item if city:canBuild(item) returns false for all items in the list
--          Be careful that the 'can build chain' does not form a loop
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .requireSomeAsAlternateProduction = unitTypeObject or improvementObject or wonderObject or table of these objects
--          the city can only build the item if the city can build some of the items in the list
--          Be careful that the 'can build chain' does not form a loop
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .numberOfAlternateProduction = integer
--          tells how many of 'someAlternateProduction' is required
--          absent means no restriction
--      .conditionFunction = function(defaultBuildFunction,city,item) --> bool
--          if function returns true, item can be built if other conditions are met, if false, item can't be built
--          absent means no extra condition
--      .returnFalse = bool or nil
--          if true, item can't be built
--          if false or nil, refer to other conditions
--          (happens before overrideFunction and alternateParameters)
--      .overrideFunction = function(defaultBuildFunction,city,item) --> boolean
--          if function returns true, the city is automatically allowed to build the item, regardless of any
--          conditions that isn't met
--          if function returns false, the other conditions are checked
--      .alternateParameters = table of itemParameters
--          itemParameters is this table of restrictions on whether a given item can be produced
--          if the item in question satisfies any of the itemParameters in the table, it can be produced,
--          regardless of whether the 'top' itemParameters are satisfied
--          use this (or overrideFunction) if you want to have more than one valid way to produce the item
--          the 'table' format is important.  Unlike other parameters, you must enclose the value of 
--          alternateParameters in a table, even if there is only one itemParameters as the value
--      .computerOnly = bool or nil
--          if true, item can only be built by computer controlled players
--          if false or nil, either human or AI players can build
--          (in conjunction with alternateParameters, this can be used to have different conditions for the
--          ai and human players)
--      .humanOnly = bool or nil
--          if true, item can only be built by human controlled players
--          if false or nil, either human or AI players can build
--          (in conjunction with alternateParameters, this can be used to have different conditions for the
--          ai and human players)
--
--      .onlyBuildCoastal = bool or nil
--          if true, the item can only be built if the city has the 'coastal' flag,
--          that is, in the default game it could build harbors and offshore platforms
--      .onlyBuildShips = bool or nil
--          if true, the item can only be built if the city has the 'ship building' flag
--          that is, in the default game it could build sea units
--      .onlyBuildHydroPlant = bool or nil
--          if true, the item can only be built if the city has the 'can build hydro plant' flag
--          that is, the city could build hydro plants in the default game
--
--

--Russian Specific Locations
local EliteRegimentCities = {{22,34,0},{25,21,0},{30,30,0},{32,36,0},{34,42,0},{36,28,0},{82,38,0},{107,51,0},{108,42,0}}

--Cossack Specific Locations
local CossackCities = {{21,41,0},{25,41,0},{25,47,0},{28,46,0},{29,41,0},{31,45,0}}

--Finn Specific Locations
local FinnHomeCities = {{18,16,0},{19,11,0},{19,21,0},{22,20,0}}

--Pole Specific Locations
local PoleHomeCities = {{16,36,0},{17,41,0}}

--Japanese Specific Locations
local JapaneseHomeCities = {{104,82,0},{115,53,0},{115,57,0},{116,62,0},{116,66,0},{113,69,0},{110,72,0},{108,70,0},{107,75,0}}

--Chinese Specific Locations
local ChineseHomeCities = {{70,56,0},{74,58,0},{78,64,0},{82,72,0},{84,66,0},{84,80,0},{85,57,0},{86,74,0},{88,84,0},{88,88,0},{89,59,0},
				{89,69,0},{90,76,0},{90,92,0},{92,72,0},{92,82,0},{92,88,0},{93,59,0},{93,63,0},{94,90,0},{95,69,0},{95,85,0},
				{96,64,0},{96,74,0},{96,82,0},{98,72,0},{98,78,0},{100,56,0},{101,51,0},{102,46,0}}

--Korean Specific Locations
local KoreanHomeCities =  {105,67,0},{102,62,0}

--Mongolian Specific Locations
local MongolHomeCities =  {{98,44},{93,43,0},{84,44,0},{73,43,0}}

--Malaysian Specific Locations
local MalayHomeCities =  {{79,115,0},{81,121,0},{84,126,0},{94,124,0}}

--German Specific Locations
local GermanHomeCities = {{4,36,0},{7,35,0},{7,39,0},{9,33,0},{10,40,0},{10,44,0},{11,37,0},{13,33,0},{13,41,0},{13,45,0},{16,32,0}}

--American Specific Locations
local USMCTrainingCities = {{226,62,0},{228,58,0},{229,55,0},{194,70,0},{190,62,0},{196,74,0}}

--Mexican Specific Locations
local MexicanCities = {{202,82,0},{207,83,0},{204,90,0},{209,89,0},{208,96,0},{208,102,0},{213,97,0},{216,92,0}}

--British Specific Locations
local ScottishHomeCities = {{283,25,0}}

--Irish Irregulars Special Condition
local IrishHomeCities = {{281,29,0},{281,33,0}}
local irregularAltCondition = {location=IrishHomeCities,allImprovements=object.iBarracks, overrideDefaultBuildFunction=true}

local IndianHomeCities = {{54,82,0},{55,75,0},{55,89,0},{58,68,0},{58,94,0},{59,73,0},{59,87,0},{61,79,0},{61,107,0},
				{64,108,0},{63,89,0},{65,83,0},{67,99,0},{68,82,0},{69,95,0},{70,88,0},{72,82,0},{73,87,0}}

--French Specific Locations
local FrenchHomeCities = {{2,42,0},{2,38,0},{6,44,0},{3,47,0},{3,51,0},{0,50,0},{286,46,0},{284,44,0}}

--North African Specific Locations
local ZouavesHomeCities =  {{8,68,0},{4,66,0},{0,68,0}}

--Turk Specific Locations
local TurkHomeCities = {{24,56,0},{27,59,0},{23,61,0},{27,63,0},{30,56,0}}

--Egyptian Specific Locations
local EgyptHomeCities =  {{23,77,0},{23,73,0}}

--Ethiopian Specific Locations
local EthiopianHomeCities =  {{13,123,0},{25,107,0},{26,120,0},{29,111,0},{32,104,0},{34,110,0},{36,124,0},{38,116,0}}

--Zulu Specific Locations
local ZuluHomeCities =  {{20,162,0},{23,169,0},{25,161,0},{26,178,0},{27,127,0},{28,156,0},{28,172,0},{29,163,0},
				{32,134,0},{32,142,0},{33,155,0},{34,138,0}}

--Boer Specific Locations
local BoerHomeCities = {{12,172,0},{14,186,0},{15,191,0},{16,170,0},{17,183,0},{21,185,0},{22,178,0},{25,185,0}}

--Balkan Specific Locations
local BalkanHomeCities =  {{17,57,0},{19,55,0},{21,57,0}}

--Engineer Special Captial Condition
local engineerAltCondition = {allImprovements=object.iCapital, overrideDefaultBuildFunction=true}

local unitTypeBuild = {}

--NAVAL UNITS
unitTypeBuild[object.uSteamYacht.id] = {allImprovements={object.iHarbour,object.iLocalIndustry}}
unitTypeBuild[object.uFreighter.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uOceanLiner.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uGunboat.id] = {allImprovements={object.iHarbour,object.iLocalIndustry}}
unitTypeBuild[object.uIronclad.id] = {allImprovements={object.iHarbour,object.iFactory}}
unitTypeBuild[object.uSteamFrigate.id] = {allImprovements={object.iHarbour,object.iLocalIndustry}}
unitTypeBuild[object.uSteamWarship.id] = {allImprovements={object.iHarbour,object.iFactory}}
unitTypeBuild[object.uDestroyer.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uLightCruiser.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uArmouredCruiser.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uBattleship.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uSubmarine.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uAeroCarrier.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uDreadnought.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}
unitTypeBuild[object.uBattlecruiser.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}

--GENERAL EQUIPMENT (NON-NAVAL)
unitTypeBuild[object.uLabourers.id] = {allImprovements={object.iLocalIndustry}}
unitTypeBuild[object.uRailWorker.id] = {allImprovements={object.iLocalIndustry}}
unitTypeBuild[object.uExpedition.id] = {allImprovements={object.iUniversity}}
unitTypeBuild[object.uAssaultTroops.id] = {allImprovements={object.iFactory,object.iBarracks}}
unitTypeBuild[object.uIrregulars.id] = {allImprovements={object.iColonialPolice}}
unitTypeBuild[object.uPeasantRifles.id] = {allImprovements={object.iBarracks}}
unitTypeBuild[object.uMilitia.id] = {allImprovements={object.iBarracks}}
unitTypeBuild[object.uMarines.id] = {allImprovements={object.iHarbour,object.iBarracks}}
unitTypeBuild[object.uMachineGuns.id] = {allImprovements={object.iFactory,object.iBarracks}}
unitTypeBuild[object.uCannon.id] = {allImprovements={object.iLocalIndustry}}
unitTypeBuild[object.uCavalry.id] = {allImprovements={object.iBarracks}}
unitTypeBuild[object.uRifledGun.id] = {allImprovements={object.iLocalIndustry,object.iBarracks}}
unitTypeBuild[object.uArtillery.id] = {allImprovements={object.iFactory,object.iBarracks}}
unitTypeBuild[object.uFieldArtillery.id] = {allImprovements={object.iSteelMill,object.iBarracks}}
unitTypeBuild[object.uHeavyArtillery.id] = {allImprovements={object.iSteelMill,object.iBarracks}}
unitTypeBuild[object.uMonoplane.id] = {allImprovements={object.iAutomobilePlant,object.iAirpark}}
unitTypeBuild[object.uBiplane.id] = {allImprovements={object.iAutomobilePlant,object.iAirpark}}
unitTypeBuild[object.uBomber.id] = {allImprovements={object.iAutomobilePlant,object.iAirpark}}
unitTypeBuild[object.uAirship.id] = {allImprovements={object.iAutomobilePlant,object.iAirpark}}
unitTypeBuild[object.uMaximGun.id] = {allImprovements={object.iFactory,object.iBarracks}}
unitTypeBuild[object.uMfgGoods.id] = {allImprovements={object.iGrandRailStation}}
unitTypeBuild[object.uGoodsTruck.id] = {allImprovements={object.iAutomobilePlant}}
unitTypeBuild[object.uArmouredCar.id] = {allImprovements={object.iAutomobilePlant,object.iBarracks}}
unitTypeBuild[object.uTank.id] = {allImprovements={object.iAutomobilePlant,object.iBarracks}}
unitTypeBuild[object.uLightTank.id] = {allImprovements={object.iAutomobilePlant,object.iBarracks}}
--An empire's capital can train Engineers, even without a university or the required tech
unitTypeBuild[object.uEngineers.id] = {allImprovements=object.iUniversity,alternateParameters= {engineerAltCondition}}
--An empire can train Irregulars in Irish cities, even without the required tech
unitTypeBuild[object.uIrregulars.id] = {allImprovements=object.iColonialPolice,alternateParameters= {irregularAltCondition}}
--Agents can only be built by an empire who controls the Diplomatic Corps
unitTypeBuild[object.uAgent.id]= {allWonders={object.wDiplomaticCorps},minimumPopulation=12,numberOfWonders=1}

--RUSSIAN UNITS
--Only the Russian homelands can train Infantry
unitTypeBuild[object.uInfantry.id] = {allImprovements={object.iBarracks}}
unitTypeBuild[object.uInfantryII.id] = {allImprovements={object.iLocalIndustry,object.iBarracks}}
unitTypeBuild[object.uInfantryIII.id] = {allImprovements={object.iFactory,object.iBarracks}}
--Only certain Russian cities with officer schools in the form of Universities can build Elite regiments
unitTypeBuild[object.uEliteRegiment.id] = {location=EliteRegimentCities,allImprovements={object.iUniversity,object.iBarracks}}
--Only certain Russian cities can build Cossack regiments
unitTypeBuild[object.uCossacks.id] = {location=CossackCities,allImprovements={object.iLocalIndustry,object.iBarracks}}
unitTypeBuild[object.uCossacksII.id] = {location=CossackCities,allImprovements={object.iLocalIndustry,object.iBarracks}}

--POLISH UNITS
--Control of the Polish home cities will enable an  empire to recruit Polish Lancers
unitTypeBuild[object.uPolishLancers.id] = {location=PoleHomeCities,allImprovements={object.iBarracks}}

--BALKAN UNITS
--Control of the Balkan cities will enable an  empire to train Balkan Rebels
unitTypeBuild[object.uBalkanRebels.id] = {location=BalkanHomeCities,allImprovements={object.iBarracks}}

-- FINNISH UNITS
--Control of the Finn  home cities will enable an  empire to train Finn Rebels
unitTypeBuild[object.uFinnishRebels.id] = {location=FinnHomeCities,allImprovements={object.iBarracks}}

--JAPANESE UNITS
--Only the Japanese home islands can train Imperial Marines and train Imperial Armies
unitTypeBuild[object.uImperialMarines.id] = {location=JapaneseHomeCities,allImprovements={object.iHarbour,object.iBarracks}}
unitTypeBuild[object.uImperialArmy.id] = {location=JapaneseHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uImperialArmyII.id] = {location=JapaneseHomeCities,allImprovements={object.iBarracks}}

--CHINESE UNITS
--Control of the Chinese  home cities will enable an  empire to train Chinese Armies
unitTypeBuild[object.uChineseArmy.id] = {location=ChineseHomeCities,allImprovements={object.iBarracks}}

--KOREAN UNITS
--Control of the Korean  home cities will enable an  empire to train Korean Raiders
unitTypeBuild[object.uBanditRaiders.id] = {location=KoreanHomeCities,allImprovements={object.iBarracks}}

--MONGOLIAN UNITS
--Control of the Mongol  home cities will enable an  empire to train Tarar Cavalry
unitTypeBuild[object.uTatarCavalry.id] = {location=MongolHomeCities,allImprovements={object.iBarracks}}

--GERMAN UNITS
--Only the German homelands can train Grenadiers
unitTypeBuild[object.uGrenadiers.id] = {location=GermanHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uGrenadiersII.id] = {location=GermanHomeCities,allImprovements={object.iLocalIndustry,object.iBarracks}}
unitTypeBuild[object.uGrenadiersIII.id] = {location=GermanHomeCities,allImprovements={object.iFactory,object.iBarracks}}
--Only the German homelands can train Stormtroopers
unitTypeBuild[object.uStormtroopers.id] = {location=GermanHomeCities,allImprovements={object.iFactory,object.iBarracks}}
--Only the German homelands can train Uhlans
unitTypeBuild[object.uUhlans.id] = {location=GermanHomeCities,allImprovements={object.iLocalIndustry,object.iBarracks}}
--Only the German homelands can build Field Guns
unitTypeBuild[object.uFieldGun.id] = {location=GermanHomeCities,allImprovements={object.iFactory,object.iSteelMill}}

--AMERICAN UNITS
unitTypeBuild[object.uUSArmy.id] = {allImprovements={object.iBarracks}}
unitTypeBuild[object.uUSArmyII.id] = {allImprovements={object.iLocalIndustry,object.iBarracks}}
unitTypeBuild[object.uUSArmyIII.id] = {allImprovements={object.iFactory,object.iBarracks}}
unitTypeBuild[object.uUSMarines.id] = {location=USMCTrainingCities,allImprovements={object.iHarbour,object.iBarracks}}
unitTypeBuild[object.uUSCavalry.id] = {allImprovements={object.iLocalIndustry,object.iBarracks}}
unitTypeBuild[object.uParrottRifle.id] = {allImprovements={object.iLocalIndustry,object.iBarracks}}

--MEXICAN UNITS
--Control of the Mexican home cities will enable an  empire to train Guerrillas
unitTypeBuild[object.uGuerrillas.id] = {location=MexicanCities,allImprovements={object.iBarracks}}

--BRITISH UNITS
unitTypeBuild[object.uRiflemen.id] = {allImprovements={object.iBarracks}}
unitTypeBuild[object.uRiflemenII.id] = {allImprovements={object.iLocalIndustry,object.iBarracks}}
unitTypeBuild[object.uRiflemenIII.id] = {allImprovements={object.iFactory,object.iBarracks}}
unitTypeBuild[object.uHeavyGun.id] = {allImprovements={object.iFactory,object.iSteelMill}}
--Only Glasgow can train Highlander regiments
unitTypeBuild[object.uHighlanders.id] =  {location=ScottishHomeCities,allImprovements={object.iBarracks}}
--Only Glasgow can train Highland Rifle regiments
unitTypeBuild[object.uHighlandRifles.id] = {location=ScottishHomeCities,allImprovements={object.iFactory,object.iBarracks}}
--Indian Troop regiments can only be trained in the British Raj
unitTypeBuild[object.uIndianTroops.id] = {location=IndianHomeCities,allImprovements={object.iBarracks}}
--Gurkha regiments can only be trained in the British Raj
unitTypeBuild[object.uGurkhas.id] = {location=IndianHomeCities,allImprovements={object.iBarracks}}
--Sepoy regiments can only be trained in the British Raj
unitTypeBuild[object.uSepoys.id] = {location=IndianHomeCities,allImprovements={object.iBarracks}}

--TURKISH UNITS
--Control of the Turkish home cities will enable an  empire to recruit Turk Cannons
unitTypeBuild[object.uTurkCannon.id] = {location=TurkHomeCities,allImprovements={object.iBarracks}}

--EGYPTIAN UNITS
--Control of the Egyptian home cities will enable an  empire to train Egyptian Armies
unitTypeBuild[object.uEgyptianArmy.id] = {location=EgyptHomeCities,allImprovements={object.iBarracks}}

--ETHIOPIAN UNITS
--Control of the Ethiopian home cities will enable an  empire to train Tribal Cavalry
unitTypeBuild[object.uTribalCavalry.id] = {location=EthiopianHomeCities,allImprovements={object.iBarracks}}

--ZULU UNITS
--Control of the Zulu home cities will enable an  empire to train Tribal Rifles
unitTypeBuild[object.uTribalRifles.id] = {location=ZuluHomeCities,allImprovements={object.iBarracks}}

--BOER UNITS
--Control of the Boer home cities will enable an  empire to train Boer Troops
unitTypeBuild[object.uBoerTroops.id] = {location=BoerHomeCities,allImprovements={object.iBarracks}}
--Control of the Boer home cities will enable an  empire to train Boer Kommandos
unitTypeBuild[object.uBoerKommando.id] = {location=BoerHomeCities,allImprovements={object.iLocalIndustry,object.iBarracks}}

--MALAYSIAN UNITS
--Control of the Malay home cities will enable an  empire to train Malay Warriors
unitTypeBuild[object.uMalayWarriors.id] = {location=MalayHomeCities,allImprovements={object.iBarracks}}

--FRENCH UNITS
unitTypeBuild[object.uInfantrie.id] = {location=FrenchHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uInfantrieII.id] = {location=FrenchHomeCities,allImprovements={object.iLocalIndustry,object.iBarracks}}
unitTypeBuild[object.uInfantrieIII.id] = {location=FrenchHomeCities,allImprovements={object.iFactory,object.iBarracks}}
unitTypeBuild[object.uChasseurs.id] = {location=FrenchHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uBulletScout.id] = {location=FrenchHomeCities,allImprovements={object.iFactory,object.iAirpark}}
--Zouaves can be trained in certain North African cities
unitTypeBuild[object.uZouaves.id] = {location=ZouavesHomeCities,allImprovements={object.iBarracks}}

local improvementBuild = {}

--Factories can be built in cities at size 6 with a Bank, and also a Grand Rail Station
improvementBuild[object.iFactory.id] = {someImprovements={object.iBank,object.iGrandRailStation},numberOfImprovements=2,minimumPopulation=6}

--Modern Housing can be built in cities at size 8 with a Theatre, and also a Grand Rail Station
improvementBuild[object.iModernHousing.id] = {someImprovements={object.iTheatre,object.iGrandRailStation},numberOfImprovements=2,minimumPopulation=8}

--Secret Police HQs can be built in cities at size 8 with Colonial Police, and a Magistrate
improvementBuild[object.iSecretPoliceHQ.id] = {someImprovements={object.iColonialPolice,object.iMagistrate},numberOfImprovements=2,minimumPopulation=8}

local wonderBuild = {}

--Diplomatic Corps Wonder can only be constructed in the capital
wonderBuild[object.wDiplomaticCorps.id] = {allImprovements=object.iCapital}

--Imperial Powers Wonder can only be constructed in the capital
wonderBuild[object.wImperialPowers.id] = {allImprovements=object.iCapital}

--Women's Suffrage Wonder can only be constructed in the capital
wonderBuild[object.wWomensSuffrage.id] = {allImprovements=object.iCapital}

-- Eugenics Foundation Wonder can only be constructed in a city with a University
wonderBuild[object.wEugenicsFoundation.id] = {allImprovements=object.iUniversity}

--International Refinery Wonder can only be constructed in a city with a Stock Exchange
wonderBuild[object.wInternationalRefinery.id] = {allImprovements=object.iStockExchange}

--Central Power Station Wonder can only be constructed in a city with a Factory
wonderBuild[object.wCentralPowerStation.id] = {allImprovements=object.iFactory}

--Naval Expansion Wonder can only be constructed in a city with an International Port and Steel Mill
wonderBuild[object.wNavalExpansion.id] = {allImprovements={object.iInternationalPort,object.iSteelMill}}

--Armament Program Wonder can only be constructed in a city with a Laboratory and Factory
wonderBuild[object.wArmamentProgram.id] = {allImprovements={object.iLaboratory,object.iFactory}}


canBuildFunctions.supplyUnitTypeParameters(unitTypeBuild)
canBuildFunctions.supplyImprovementParameters(improvementBuild)
canBuildFunctions.supplyWonderParameters(wonderBuild)
